#!/usr/bin/perl
use strict;
use warnings 'FATAL' => 'all';
use XML::Simple ();

my $xml = XML::Simple::XMLin('-', 'KeyAttr' => [], 'ForceArray' => 1);
die "Usage: $0 < <map.xml> > <polys.txt>\n" unless $xml;
our $map = $xml->{'level'};
die "Usage: $0 < <map.xml> > <polys.txt>\n" unless $map;


# Sector types:
# Void,				// Inaccessible
# Normal,				// Accessible, but nothing special
# Door,				// (self-explanatory)
# ChangeLevel,		// Change levels here
# DoorTrigger,		// Triggers doors opening/closing
# SecretDoor,			// What triggers them?
# Corpse,				// You can talk to these
# Pillar,				// Pillar in the middle -- can't walk through
# OtherTrigger,		// Not exactly sure what this one does
# Save,				// Can save game here


for my $level (0..24)
{
  my $mp = $map->[$level];
  unless ($mp)
  {
#     warn "Failed to find level info for $level\n";
    last;
  }

  my $doorlist = $mp->{'doors'}[0]{'door'};
  my $ladderlist = $mp->{'level_changes'}[0]{'level_change'};
  my $sectorlist = $mp->{'sectors'}[0]{'sector'};
  
  # initialize map
  my @grid;
  for my $row (0..33)
  {
    $grid[$row] = [];
    for my $col (0..33)
    {
      $grid[$row][$col] = [ 0, 0, [ 0, 0, 0, 0 ] ];
    }
  }
  
  # deal with non-void sectors
  for my $sector (@$sectorlist)
  {
    my $type = $sector->{'type'};
    my $col = $sector->{'col'} + 1;
    my $row = $sector->{'row'} + 1;
    
    my $bl = ($sector->{'corner_bl_type'} || 0) == 160 ? 1 : 0;
    my $tl = ($sector->{'corner_tl_type'} || 0) == 160 ? 1 : 0;
    my $tr = ($sector->{'corner_tr_type'} || 0) == 160 ? 1 : 0;
    my $br = ($sector->{'corner_br_type'} || 0) == 160 ? 1 : 0;
    
    my $subtype = 0;
    if ($type == 2) # door
    {
      my $door = IndexOf($doorlist, $sector->{'extra'});
      die "Door not found at $level ($col, $row)" unless $door;
      $subtype = $door->{'direction'};
    }
      
    $grid[$row][$col] = [ $type, $subtype, [ $bl, $tl, $tr, $br ] ];
    
  }
  
  # detect unreachable void sectors
  my @unreachable;
  for my $row (1..32)
  {
    for my $col (1..32)
    {
      my $gref = $grid[$row][$col];
      next unless $gref->[0] == 0;
#       next if ($gref->[2][0] || $gref->[2][1] || $gref->[2][2] || $gref->[2][3]);
      
      next unless
        $grid[$row - 1][$col - 1][0] == 0 &&
        $grid[$row + 0][$col - 1][0] == 0 &&
        $grid[$row + 1][$col - 1][0] == 0 &&
        $grid[$row - 1][$col + 0][0] == 0 &&
        $grid[$row + 1][$col + 0][0] == 0 &&
        $grid[$row - 1][$col + 1][0] == 0 &&
        $grid[$row + 0][$col + 1][0] == 0 &&
        $grid[$row + 1][$col + 1][0] == 0;
      
      push(@unreachable, [ $row, $col ]);
    }
  }
  
  # mark borders as unreachable
  for my $row (0..33)
  {
    for my $col (0, 33)
    {
      push(@unreachable, [ $row, $col ]);
    }
  }
  for my $col (1..32)
  {
    for my $row (0, 33)
    {
      push(@unreachable, [ $row, $col ]);
    }
  }
  
  # change unreachable to open, to mimic map behavior
  for my $ref (@unreachable)
  {
    my ($row, $col) = @$ref;
    $grid[$row][$col][0] = 1;
    $grid[$row][$col][1] = 999;
  }
  
  # now, we can actually deal with sectors
  my $prefix = sprintf("%02d %+07.1f", $level, $mp->{'height'});
  my @cspecs = (
    [ (0, 0), (0.25, 0), (0, 0.25) ],
    [ (0, 1), (0, 0.75), (0.25, 1) ],
    [ (1, 1), (0.75, 1), (1, 0.75) ],
    [ (1, 0), (1, 0.25), (0.75, 0) ],
    );
  my @dspecs = (
    [ (0, 0.25), (0, 0.75), (0.5, 0.75), (0.5, 0.25) ],
    [ (0.25, 1), (0.75, 1), (0.75, 0.5), (0.25, 0.5) ],
    [ (0.5, 0.25), (0.5, 0.75), (1, 0.75), (1, 0.25) ],
    [ (0.25, 0.5), (0.75, 0.5), (0.75, 0), (0.25, 0) ],
    );
  my @pspecs = (
    [ (0.25, 0.5), (0.5, 0.75), (0.75, 0.5), (0.5, 0.25) ],
    );
  
  for my $row (0..31)
  {
    for my $col (0..31)
    {
      my $gref = $grid[$row + 1][$col + 1];
      my ($type, $subtype, $cref) = @$gref;
      next if $subtype == 999; # unreachable

      PrintPoly($prefix, 'floor',
                Vertex($col + $cspecs[0][0], $row + $cspecs[0][1]),
                Vertex($col + $cspecs[1][0], $row + $cspecs[1][1]),
                Vertex($col + $cspecs[2][0], $row + $cspecs[2][1]),
                Vertex($col + $cspecs[3][0], $row + $cspecs[3][1]));          
      
      if ($type == 0)
      {
        my @vs;
        for my $corner (0..3)
        {
          my $cc = $cspecs[$corner];
          if ($cref->[$corner])
          {
            push(@vs, Vertex($col + $cc->[2], $row + $cc->[3]),
                      Vertex($col + $cc->[4], $row + $cc->[5]));
          }
          else
          {
            push(@vs, Vertex($col + $cc->[0], $row + $cc->[1]));
          }
        }
        PrintPoly($prefix, 'solid', @vs);
      }
      else
      {
        for my $corner (0..3)
        {
          if ($cref->[$corner])
          {
            my $cc = $cspecs[$corner];
            PrintPoly($prefix, 'solid',
                      Vertex($col + $cc->[0], $row + $cc->[1]),
                      Vertex($col + $cc->[4], $row + $cc->[5]),
                      Vertex($col + $cc->[2], $row + $cc->[3]));
          }
        }
        
        if ($type == 2)
        {
          # it's a door
          my $dd = $dspecs[$subtype];
          PrintPoly($prefix, 'door',
                    Vertex($col + $dd->[0], $row + $dd->[1]),
                    Vertex($col + $dd->[2], $row + $dd->[3]),
                    Vertex($col + $dd->[4], $row + $dd->[5]),
                    Vertex($col + $dd->[6], $row + $dd->[7]));
        }
        elsif ($type == 7)
        {
          # it's a pillar
          my $pp = $pspecs[$subtype];
          PrintPoly($prefix, 'pillar',
                    Vertex($col + $pp->[0], $row + $pp->[1]),
                    Vertex($col + $pp->[2], $row + $pp->[3]),
                    Vertex($col + $pp->[4], $row + $pp->[5]),
                    Vertex($col + $pp->[6], $row + $pp->[7]));
        }
      }
    }
  }
}

sub IndexOf
{
  my ($ref, $idx) = @_;
  
  for my $r (@$ref)
  {
    if ($r->{'index'} == $idx)
    {
      return $r;
    }
  }
  return undef;
}

sub Vertex
{
  my ($x, $y) = @_;
  
  return sprintf('%05.2f,%05.2f', $x, $y);
}

sub PrintPoly
{
  my ($prefix, $txt, @vs) = @_;
  
  return unless scalar @vs;
  print join(" ", $prefix, $txt, @vs) . "\n";
}
